from ..util import reify

from .base_resource import Resource
from .service import Service


@Resource.register_resource('Conformance')
class Conformance(Resource):

    @property
    def rest(self):
        return [
            Service(it, _bind=self.bind)
            for it in self['rest']]


@Resource.register_resource('CapabilityStatement')
class CapabilityStatement(Resource):

    @property
    def rest(self):
        return [
            Service(it, _bind=self.bind)
            for it in self['rest']]




from functools import update_wrapper


class AttrProxy:

    def __init__(self, obj, attr):
        self._obj = obj
        self._attr = attr

    def __getattr__(self, name):
        attr = getattr(self._obj, self._attr)
        return getattr(attr, name)


def reify(func):
    result = _Reified(func)
    update_wrapper(result, func)
    return result


class _Reified(object):

    def __init__(self, func, name=None):
        self._func = func
        if name is None:
            name = func.__name__
        self._name = name

    def __get__(self, obj, cls=None):
        if obj is None:
            return None
        result = obj.__dict__[self._name] = self._func(obj)
        return result



from urllib.parse import urlencode

class Resource(dict):
    _registry = {}
    metadata = {}
    resourceType = ''

    def __init__(self, *args, **kwargs):
        self.bind = kwargs.pop('_bind', None)
        if self.bind is None:
            self.bind = self.metadata.get('bind')
        super().__init__(*args, **kwargs)

    def __repr__(self):
        bound = 'Bound' if self.bind else 'Unbound'
        if self.resourceType:
            return f'<{bound} {self.resourceType}>'
        else:
            return super().__repr__()

    @classmethod
    def search(cls, spec, bind=None):
        if bind is None:
            bind = cls.metadata.get('bind')
        qs = urlencode(spec, doseq=True)
        resp = bind.get(f'/{cls.resourceType}?{qs}')
        return cls.from_dict(resp, bind=bind)

    @classmethod
    def fetch(cls, id, bind=None):
        if bind is None:
            bind = cls.metadata.get('bind')
        resp = bind.get(f'/{cls.resourceType}/{id}')
        return cls.from_dict(resp, bind=bind)

    @classmethod
    def register_resource(cls, resourceType):
        def decorator(subclass):
            subclass.resourceType = resourceType
            cls._registry[resourceType] = subclass
            return subclass
        return decorator

    @classmethod
    def resourceClass(cls, name, **metadata):
        result = cls._registry.get(name)
        if result is not None:
            return result
        result = cls._registry[name] = type(
            name, (Resource,),
            {'resourceType': name, 'metadata': metadata})
        return result

    @classmethod
    def from_dict(cls, dct, resourceType=None, bind=None):
        if resourceType is None:
            resourceType = dct.get('resourceType', '')
        dct = {
            key: _make_resources(value, bind=bind)
            for key, value in dct.items()}
        resourceClass = cls.resourceClass(resourceType)
        return resourceClass(dct, _bind=bind)

    def resolve(self):
        return self.bind.get(f'/{self.reference}')

    def __dir__(self):
        return [*super().__dir__(), *self.keys()]

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)


def _make_resources(obj, bind=None):
    if isinstance(obj, list):
        return [_make_resources(item, bind=bind) for item in obj]
    elif isinstance(obj, dict):
        return Resource.from_dict(obj, bind=bind)
    else:
        return obj



import copy
from .util import reify

import requests
from requests_oauthlib import OAuth2Session
from .resource import Resource


class Connection:

    def __init__(self, service_root, session=None, client_secret=None):
        self.service_root = service_root.rstrip('/')
        if session is None:
            session = OAuth2Session()
        self.session = session
        self.client_secret = client_secret
        self.session.headers['Accept'] = 'application/json'

    def __repr__(self):
        return f'<Connection {self.service_root}>'

    @classmethod
    def from_token(cls, service_root, token):
        session = OAuth2Session(token=token)
        self = cls(service_root, session=session)

    @reify
    def metadata(self):
        url = self.service_root + '/metadata'
        resp = requests.get(url, headers={'Accept': 'application/json'})
        resp.raise_for_status()
        return resp.json()

    def service(self, rest_id=0):
        return Resource.from_dict(
            self.metadata['rest'][rest_id],
            resourceType='_Service',
            bind=self)

    def get(self, path):
        url = self.service_root + path
        resp = self.session.get(url)
        resp.raise_for_status()
        return resp.json()

    def clone(self, **kwargs):
        result = copy.copy(self)
        for k, v in kwargs.items():
            setattr(result, k, v)
        return result



from setuptools import setup, find_packages

setup(
    name='FHIRstorm',
    # version='0.0.0',
    version='0.0.1',
    version_format='{tag}.dev{commitcount}',
    setup_requires=['setuptools-git-version'],
    description='SMART on FHIR client for Python',
    long_description='',
    classifiers=[
        "Programming Language :: Python",
    ],
    author='Rick Copeland',
    author_email='rick@arborian.com',
    url='',
    keywords='',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[],
    tests_require=[],
    entry_points="""
    """)



import logging
from urllib.parse import urlparse, parse_qs
from datetime import datetime, timedelta

import jwt
from requests_oauthlib import OAuth2Session

log = logging.getLogger(__name__)


def jwt_state(secret, expires_in=300, **claims):
    claims = dict(claims)
    claims.setdefault('exp', datetime.utcnow() + timedelta(seconds=expires_in))
    return jwt.encode(claims, secret)


def jwt_state_validator(secret, **kwargs):
    def validator(state):
        return jwt.decode(state, secret, **kwargs)
    return validator


def authorization_url(
        service, client_id, redirect_uri, scope, state,
        client_secret=None,
        aud=None,
        launch=None):
    conn = service.bind.clone(
        session=OAuth2Session(
            client_id=client_id,
            redirect_uri=redirect_uri,
            scope=scope))
    return conn.session.authorization_url(
        service.security.oauth2_uris.authorize,
        client_secret=client_secret,
        aud=aud,
        state=state,
        launch=launch)


def fetch_token(
        service, client_id, redirect_uri,
        authorization_response,
        client_secret=None,
        state=None,
        state_validator=None):
    if state_validator:
        pr = urlparse(authorization_response)
        state_validator(parse_qs(pr.query)['state'][0])
    conn = service.bind.clone(
        session=OAuth2Session(
            client_id=client_id,
            redirect_uri=redirect_uri))
    return conn.session.fetch_token(
        token_url=service.security.oauth2_uris.token,
        authorization_response=authorization_response,
        client_id=client_id,
        client_secret=client_secret,
        state=state)



from .base_resource import Resource

from ..util import AttrProxy


@Resource.register_resource('_Service')
class Service(Resource):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._registry = {}
        self.r = ResourceCollection(self)

    @property
    def security(self):
        return Security(self['security'], _bind=self.bind)

    def resourceClass(self, name, **metadata):
        result = self._registry.get(name)
        if result is not None:
            return result
        base_cls = Resource.resourceClass(name)
        result = self._registry[name] = type(
            name, (base_cls,), {'metadata': metadata})
        return result


    def refresh_token(self):
        if self.bind.client_secret:
            auth = (self.bind.session.client_id, self.bind.client_secret)
        else:
            auth = None
        new_token = self.bind.session.refresh_token(
            self.security.oauth2_uris.token,
            auth=auth)
        return new_token


class ResourceCollection:

    def __init__(self, service):
        self._service = service
        self._resourceTypes = [
            res['type'] for res in self._service.resource]
        for res in self._service.resource:
            setattr(self, res['type'], self._service.resourceClass(
                res['type'],
                bind=AttrProxy(self._service, 'bind'),
                **res))

    def __iter__(self):
        return iter(self._resourceTypes)

    def items(self):
        return (
            (t, getattr(self, t)) for t in self)

    def __dir__(self):
        return list(self)


@Resource.register_resource('_Security')
class Security(Resource):

    @property
    def oauth2_uris(self):
        for ext in self.extension:
            if ext['url'] == 'http://fhir-registry.smarthealthit.org/StructureDefinition/oauth-uris':
                return Oauth2Uris(
                    {e.url: e.valueUri for e in ext.extension},
                    _bind=self.bind)


class Oauth2Uris(Resource):
    resourceType = '_Oauth2Uris'
